@@ -15,6 +15,7 @@ from fcache.cache import FileCache  # type: ignore
 from sentry_sdk.integrations.aiohttp import AioHttpIntegration
 from sentry_sdk.integrations.logging import LoggingIntegration
 
+import dipdup.context as context
 from dipdup import __spec_version__, __version__, spec_reindex_mapping, spec_version_mapping
 from dipdup.codegen import DEFAULT_DOCKER_ENV_FILE, DEFAULT_DOCKER_IMAGE, DEFAULT_DOCKER_TAG, DipDupCodeGenerator
 from dipdup.config import DipDupConfig, LoggingConfig, PostgresDatabaseConfig
@@ -92,6 +93,7 @@ def init_sentry(config: DipDupConfig) -> None:
 @click.pass_context
 @cli_wrapper
 async def cli(ctx, config: List[str], env_file: List[str], logging_config: str):
+    # NOTE: Config from cwd, fallback to builtin
     try:
         path = join(os.getcwd(), logging_config)
         _logging_config = LoggingConfig.load(path)
@@ -111,6 +113,8 @@ async def cli(ctx, config: List[str], env_file: List[str], logging_config: str):
     _config = DipDupConfig.load(config)
     init_sentry(_config)
 
+    await DipDupCodeGenerator(_config, {}).create_package()
+
     if _config.spec_version not in spec_version_mapping:
         raise ConfigurationError(f'Unknown `spec_version`, correct ones: {", ".join(spec_version_mapping)}')
     if _config.spec_version != __spec_version__ and ctx.invoked_subcommand != 'migrate':
@@ -132,14 +136,24 @@ async def cli(ctx, config: List[str], env_file: List[str], logging_config: str):
 @cli.command(help='Run indexing')
 @click.option('--reindex', is_flag=True, help='Drop database and start indexing from scratch')
 @click.option('--oneshot', is_flag=True, help='Synchronize indexes wia REST and exit without starting WS connection')
+@click.option('--postpone-jobs', is_flag=True, help='Do not start job scheduler until all indexes are synchronized')
+@click.option('--forbid-reindexing', is_flag=True, help='Raise exception instead of truncating database when reindexing is triggered')
 @click.pass_context
 @cli_wrapper
-async def run(ctx, reindex: bool, oneshot: bool) -> None:
+async def run(
+    ctx,
+    reindex: bool,
+    oneshot: bool,
+    postpone_jobs: bool,
+    forbid_reindexing: bool,
+) -> None:
     config: DipDupConfig = ctx.obj.config
     config.initialize()
     set_decimal_context(config.package)
+    if forbid_reindexing:
+        context.forbid_reindexing = True
     dipdup = DipDup(config)
-    await dipdup.run(reindex, oneshot)
+    await dipdup.run(reindex, oneshot, postpone_jobs)
 
 
 @cli.command(help='Generate missing callbacks and types')
@@ -148,7 +162,7 @@ async def run(ctx, reindex: bool, oneshot: bool) -> None:
 @cli_wrapper
 async def init(ctx, overwrite_types: bool):
     config: DipDupConfig = ctx.obj.config
-    config.pre_initialize()
+    config.initialize(skip_imports=True)
     dipdup = DipDup(config)
     await dipdup.init(overwrite_types)
 
@@ -158,7 +172,7 @@ async def init(ctx, overwrite_types: bool):
 @cli_wrapper
 async def migrate(ctx):
     config: DipDupConfig = ctx.obj.config
-    config.pre_initialize()
+    config.initialize(skip_imports=True)
     migrations = DipDupMigrationManager(config, ctx.obj.config_paths)
     await migrations.migrate()
 
