@@ -15,6 +15,13 @@ from fcache.cache import FileCache  # type: ignore
 from dipdup import __version__
 from dipdup.config import HTTPConfig  # type: ignore
 
+safe_exceptions = (
+    aiohttp.ClientConnectionError,
+    aiohttp.ClientConnectorError,
+    aiohttp.ClientResponseError,
+    aiohttp.ClientPayloadError,
+)
+
 
 class HTTPGateway(ABC):
     """Base class for datasources which connect to remote HTTP endpoints"""
@@ -65,6 +72,7 @@ class _HTTPGateway:
         """Create underlying aiohttp session"""
         self.__session = aiohttp.ClientSession(
             connector=aiohttp.TCPConnector(limit=self._config.connection_limit or 100),
+            timeout=aiohttp.ClientTimeout(connect=self._config.connection_timeout or 60),
         )
 
     async def __aexit__(self, exc_type, exc, tb):
@@ -107,12 +115,12 @@ class _HTTPGateway:
                     weight=weight,
                     **kwargs,
                 )
-            except (aiohttp.ClientConnectionError, aiohttp.ClientConnectorError, aiohttp.ClientResponseError) as e:
+            except safe_exceptions as e:
                 if self._config.retry_count and attempt - 1 == self._config.retry_count:
                     raise e
 
                 ratelimit_sleep: Optional[float] = None
-                if isinstance(e, aiohttp.ClientResponseError) and e.code == HTTPStatus.TOO_MANY_REQUESTS:
+                if isinstance(e, aiohttp.ClientResponseError) and e.status == HTTPStatus.TOO_MANY_REQUESTS:
                     # NOTE: Sleep at least 5 seconds on ratelimit
                     ratelimit_sleep = 5
                     # TODO: Parse Retry-After in UTC date format
