@@ -1,6 +1,5 @@
 from abc import abstractmethod
 from collections import defaultdict, deque, namedtuple
-from contextlib import suppress
 from typing import Deque, Dict, List, Optional, Set, Tuple, Union, cast
 
 from pydantic.error_wrappers import ValidationError
@@ -20,16 +19,16 @@ from dipdup.config import (
 )
 from dipdup.context import DipDupContext
 from dipdup.datasources.tzkt.datasource import BigMapFetcher, OperationFetcher, TzktDatasource
-from dipdup.exceptions import ConfigInitializationException, InvalidDataError
-from dipdup.models import BigMapData, BigMapDiff, HeadBlockData
-from dipdup.models import Index as IndexState
-from dipdup.models import IndexStatus, OperationData, Origination, Transaction
+from dipdup.exceptions import ConfigInitializationException, InvalidDataError, ReindexingReason
+from dipdup.models import BigMapData, BigMapDiff, BlockData, IndexStatus, OperationData, Origination, Transaction
 from dipdup.utils import FormattedLogger
 from dipdup.utils.database import in_global_transaction
 
 # NOTE: Operations of a single contract call
 OperationSubgroup = namedtuple('OperationSubgroup', ('hash', 'counter'))
 
+_cached_blocks: Dict[int, BlockData] = {}
+
 
 class Index:
     _queue: Deque
@@ -52,30 +51,32 @@ class Index:
             raise RuntimeError('Index state is not initialized')
         return self._state
 
-    async def initialize_state(self, state: Optional[IndexState] = None) -> None:
-        if not self._state:
-            if not state:
-                state, _ = await models.Index.get_or_create(
-                    name=self._config.name,
-                    type=self._config.kind,
-                    defaults=dict(
-                        level=self._config.first_level,
-                        config_hash=self._config.hash(),
-                        template=self._config.parent.name if self._config.parent else None,
-                        template_values=self._config.template_values,
-                    ),
-                )
+    async def initialize_state(self) -> None:
+        if self._state:
+            raise RuntimeError('Index state is already initialized')
 
-            self._state = state
+        self._state, created = await models.Index.get_or_create(
+            name=self._config.name,
+            type=self._config.kind,
+            defaults=dict(
+                level=self._config.first_level,
+                config_hash=self._config.hash(),
+                template=self._config.parent.name if self._config.parent else None,
+                template_values=self._config.template_values,
+            ),
+        )
 
-        # NOTE: No need to check hashes of indexes which are not synchronized.
-        head = await self.state.head
-        if not head or self.state.level != head.level:
+        if created or not self._state.level:
             return
 
-        block = await self._datasource.get_block(self.state.level)
-        if head.hash != block.hash:
-            await self._ctx.reindex(reason='block hash mismatch (missed rollback while DipDup was stopped)')
+        head = await models.Head.filter(name=self.datasource.name).order_by('-level').first()
+        if not head:
+            return
+
+        if head.level not in _cached_blocks:
+            _cached_blocks[head.level] = await self.datasource.get_block(head.level)
+        if head.hash != _cached_blocks[head.level].hash:
+            await self._ctx.reindex(ReindexingReason.BLOCK_HASH_MISMATCH)
 
     async def process(self) -> None:
         # NOTE: `--oneshot` flag implied
@@ -84,16 +85,10 @@ class Index:
             await self._synchronize(last_level, cache=True)
             await self.state.update_status(IndexStatus.ONESHOT, last_level)
 
-        elif self._datasource.sync_level is None:
-            self._logger.info('Datasource is not active, sync to the latest block')
-            # NOTE: Late establishing connection to the WebSocket
-            if self.datasource.head:
-                last_level = self.datasource.head.level
-            else:
-                last_level = (await self._datasource.get_head_block()).level
-            await self._synchronize(last_level)
+        if self._datasource.sync_level is None:
+            raise RuntimeError('Call `set_sync_level` before starting IndexDispatcher')
 
-        elif self._datasource.sync_level > self.state.level:
+        elif self.state.level < self._datasource.sync_level:
             self._logger.info('Index is behind datasource, sync to datasource level')
             self._queue.clear()
             last_level = self._datasource.sync_level
@@ -111,20 +106,23 @@ class Index:
         ...
 
     async def _enter_sync_state(self, last_level: int) -> Optional[int]:
+        if self.state.status == IndexStatus.ONESHOT:
+            return None
+
         first_level = self.state.level
+
         if first_level == last_level:
             return None
         if first_level > last_level:
             raise RuntimeError(f'Attempt to synchronize index from level {first_level} to level {last_level}')
 
         self._logger.info('Synchronizing index to level %s', last_level)
-        await self.state.update_status(IndexStatus.SYNCING, first_level)
+        await self.state.update_status(status=IndexStatus.SYNCING, level=first_level)
         return first_level
 
     async def _exit_sync_state(self, last_level: int) -> None:
         self._logger.info('Index is synchronized to level %s', last_level)
-        # NOTE: No head yet, wait for realtime messages to be processed
-        await self.state.update_status(IndexStatus.REALTIME, last_level)
+        await self.state.update_status(status=IndexStatus.REALTIME, level=last_level)
 
 
 class OperationIndex(Index):
@@ -132,14 +130,14 @@ class OperationIndex(Index):
 
     def __init__(self, ctx: DipDupContext, config: OperationIndexConfig, datasource: TzktDatasource) -> None:
         super().__init__(ctx, config, datasource)
-        self._queue: Deque[Tuple[int, List[OperationData], Optional[HeadBlockData]]] = deque()
+        self._queue: Deque[Tuple[int, List[OperationData]]] = deque()
         self._contract_hashes: Dict[str, Tuple[int, int]] = {}
         self._rollback_level: Optional[int] = None
         self._last_hashes: Set[str] = set()
         self._migration_originations: Optional[Dict[str, OperationData]] = None
 
-    def push(self, level: int, operations: List[OperationData], block: Optional[HeadBlockData] = None) -> None:
-        self._queue.append((level, operations, block))
+    def push(self, level: int, operations: List[OperationData]) -> None:
+        self._queue.append((level, operations))
 
     async def single_level_rollback(self, from_level: int) -> None:
         """Ensure next arrived block is the same as rolled back one
@@ -159,13 +157,11 @@ class OperationIndex(Index):
 
     async def _process_queue(self) -> None:
         """Process WebSocket queue"""
-        if not self._queue:
-            return
-        self._logger.info('Processing websocket queue')
-        with suppress(IndexError):
-            while True:
-                level, operations, block = self._queue.popleft()
-                await self._process_level_operations(level, operations, block)
+        if self._queue:
+            self._logger.info('Processing websocket queue')
+        while self._queue:
+            level, operations = self._queue.popleft()
+            await self._process_level_operations(level, operations)
 
     async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
@@ -199,7 +195,7 @@ class OperationIndex(Index):
 
         await self._exit_sync_state(last_level)
 
-    async def _process_level_operations(self, level: int, operations: List[OperationData], block: Optional[HeadBlockData] = None) -> None:
+    async def _process_level_operations(self, level: int, operations: List[OperationData]) -> None:
         if level <= self.state.level:
             raise RuntimeError(f'Level of operation batch must be higher than index state level: {level} <= {self.state.level}')
 
@@ -208,7 +204,6 @@ class OperationIndex(Index):
                 'operations': level,
                 'rollback': self._rollback_level,
                 'index': self.state.level,
-                'block': block.level if block else level,
             }
             if len(set(levels.values())) != 1:
                 levels_repr = ', '.join(f'{k}={v}' for k, v in levels.items())
@@ -219,7 +214,7 @@ class OperationIndex(Index):
             received_hashes = set([op.hash for op in operations])
             reused_hashes = received_hashes & expected_hashes
             if reused_hashes != expected_hashes:
-                await self._ctx.reindex(reason='attempted a single level rollback, but arrived block has additional transactions')
+                await self._ctx.reindex(ReindexingReason.ROLLBACK)
 
             self._rollback_level = None
             self._last_hashes = set()
@@ -231,12 +226,7 @@ class OperationIndex(Index):
         async with in_global_transaction():
             self._logger.info('Processing %s operations of level %s', len(operations), level)
             await self._process_operations(operations)
-
-            if block:
-                status, head = IndexStatus.REALTIME, self.datasource.head
-            else:
-                status, head = IndexStatus.SYNCING, None
-            await self.state.update_status(status, level, head)
+            await self.state.update_status(self.state.status, level)
 
     async def _match_operation(self, pattern_config: OperationHandlerPatternConfigT, operation: OperationData) -> bool:
         """Match single operation with pattern"""
@@ -415,20 +405,18 @@ class BigMapIndex(Index):
 
     def __init__(self, ctx: DipDupContext, config: BigMapIndexConfig, datasource: TzktDatasource) -> None:
         super().__init__(ctx, config, datasource)
-        self._queue: Deque[Tuple[int, List[BigMapData], Optional[HeadBlockData]]] = deque()
+        self._queue: Deque[Tuple[int, List[BigMapData]]] = deque()
 
-    def push(self, level: int, big_maps: List[BigMapData], block: Optional[HeadBlockData] = None):
-        self._queue.append((level, big_maps, block))
+    def push(self, level: int, big_maps: List[BigMapData]):
+        self._queue.append((level, big_maps))
 
     async def _process_queue(self) -> None:
         """Process WebSocket queue"""
-        if not self._queue:
-            return
-        self._logger.info('Processing websocket queue')
-        with suppress(IndexError):
-            while True:
-                level, big_maps, block = self._queue.popleft()
-                await self._process_level_big_maps(level, big_maps, block)
+        if self._queue:
+            self._logger.info('Processing websocket queue')
+        while self._queue:
+            level, big_maps = self._queue.popleft()
+            await self._process_level_big_maps(level, big_maps)
 
     async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
@@ -455,19 +443,14 @@ class BigMapIndex(Index):
 
         await self._exit_sync_state(last_level)
 
-    async def _process_level_big_maps(self, level: int, big_maps: List[BigMapData], block: Optional[HeadBlockData] = None):
+    async def _process_level_big_maps(self, level: int, big_maps: List[BigMapData]):
         if level <= self.state.level:
             raise RuntimeError(f'Level of big map batch must be higher than index state level: {level} <= {self.state.level}')
 
         async with in_global_transaction():
             self._logger.info('Processing %s big map diffs of level %s', len(big_maps), level)
             await self._process_big_maps(big_maps)
-
-            if block:
-                status, head = IndexStatus.REALTIME, self.datasource.head
-            else:
-                status, head = IndexStatus.SYNCING, None
-            await self.state.update_status(status, level, head)
+            await self.state.update_status(level=level)
 
     async def _match_big_map(self, handler_config: BigMapHandlerConfig, big_map: BigMapData) -> bool:
         """Match single big map diff with pattern"""
